"""High-level UDP stream transport for MyrtIO devices"""
import asyncio
from typing import Optional

from ..message import Message, parse_message
from ..transport import MyrtIOTransport
from .persistent_stream import PersistentDatagramStream


class UDPTransport (MyrtIOTransport):
    """High-level UDP stream transport for MyrtDesk"""
    _stream: PersistentDatagramStream
    _timeout: int

    def __init__(self, host: str, port: int, timeout: int = 2):
        self._stream = PersistentDatagramStream((host, port))
        self._timeout = timeout

    async def connected(self) -> None:
        """Connects to host"""
        if not self._stream.connected:
            await self._stream.connect()

    async def run_action(self, message: Message) -> Optional[Message]:
        """Runs action"""
        try:
            async with asyncio.timeout(self._timeout):
                success = await self._stream.send(message.format_bytes())
                if not success:
                    return None
                resp = await self._stream.read()
                if resp is None:
                    return None
                return parse_message(resp)
        except asyncio.TimeoutError:
            await self._stream.reconnect()
            return None

    def close(self) -> None:
        """Closes stream"""
        self._stream.close()

async def connect_udp(host: str, port: int) -> MyrtIOTransport:
    """Creates UDP stream transport"""
    transport = UDPTransport(host, port)
    await transport.connected()
    return transport



"""MyrtIO message tests"""

from myrtio import Message, parse_message
from myrtio.constants import (
    ERROR_CODE,
    FIRST_HEADER_CODE,
    SECOND_HEADER_CODE,
    SUCCESS_CODE,
    TAIL_CODE,
)


def test_message_bytes():
    """Test message to bytes conversion"""
    feature_code = 1
    action_code = 2
    payload = [3, 4, 5]

    msg = Message(
        feature=feature_code,
        action=action_code,
        payload=payload,
    )

    assert msg.format_bytes() == bytes([
        FIRST_HEADER_CODE,
        SECOND_HEADER_CODE,
        len(payload)+2,
        feature_code,
        action_code,
        payload[0],
        payload[1],
        payload[2],
        TAIL_CODE
    ])

def test_parse_message():
    """Test message parsing"""
    msg = parse_message([
        FIRST_HEADER_CODE,
        SECOND_HEADER_CODE,
        5,
        1,
        2,
        3,
        4,
        5,
        TAIL_CODE
    ])

    assert msg.feature == 1
    assert msg.action == 2
    assert msg.payload == bytes([3, 4, 5])

def test_status():
    """Test message status"""
    msg = parse_message([
        FIRST_HEADER_CODE,
        SECOND_HEADER_CODE,
        3,
        0,
        0,
        SUCCESS_CODE,
        TAIL_CODE
    ])
    assert msg.is_successful

    msg = parse_message([
        FIRST_HEADER_CODE,
        SECOND_HEADER_CODE,
        3,
        0,
        0,
        ERROR_CODE,
        TAIL_CODE
    ])
    assert not msg.is_successful




"""MyrtIO transport interface"""
from .message import Message


class MyrtIOTransport:
    """MyrtIO transport interface"""

    async def run_action(self, message: Message) -> Message:
        """Runs action"""
        raise NotImplementedError()

    def close(self) -> None:
        """Closes transport"""
        raise NotImplementedError()



"""MyrtIO protocol constants"""

META_HEADER_PADDING = 3
MIN_MESSAGE_LENGTH = META_HEADER_PADDING + 3
MAX_MESSAGE_LENGTH = 128
# Max payload length is MAX_MESSAGE_LENGTH - header - footer
MAX_PAYLOAD_LENGTH = MAX_MESSAGE_LENGTH - META_HEADER_PADDING - 1

FIRST_HEADER_CODE = 0xFE
SECOND_HEADER_CODE = 0xEF
TAIL_CODE = 0xAF
SUCCESS_CODE = 0xEE
ERROR_CODE = 0xFF



"""Persistent datagram stream"""
import logging
from typing import Optional

from asyncio_datagram import DatagramClient, TransportClosed, connect

Address = tuple[str, int]

class PersistentDatagramStream:
    """Persistent datagram stream"""
    _host_addr: Address
    _peer_addr: Optional[Address] = None
    _stream: Optional[DatagramClient] = None

    def __init__(self, addr: Address):
        self._host_addr = addr

    async def connect(self):
        """Connects to host"""
        logging.debug("PersistentDatagramStream: connecting to %s", self._host_addr)
        if self._stream is None:
            self._stream = await connect(
                self._host_addr,
                local_addr=self._peer_addr,
                reuse_port=True)
            if self._peer_addr is None:
                self._peer_addr = self._stream.sockname
        logging.debug("PersistentDatagramStream: connected to %s", self._host_addr)

    def close(self):
        """Closes stream"""
        logging.debug("PersistentDatagramStream: close")
        if self._stream is None:
            return
        self._stream.close()
        self._stream = None

    async def reconnect(self):
        """Reconnects to host"""
        if self.connected:
            self.close()
        await self.connect()

    @property
    def port(self) -> int:
        """Returns stream host port."""
        _, port = self._peer_addr
        return port

    @property
    def addr(self) -> Address:
        """Returns stream host address."""
        return self._host_addr

    @property
    def connected(self) -> bool:
        """Returns true if stream is connected"""
        return self._stream is not None

    async def read(self) -> Optional[list[int]]:
        """Reads message from stream"""
        if self._stream is None:
            return None
        try:
            data, _ = await self._stream.recv()
            logging.debug("PersistentDatagramStream: got data (%s)", list(data))
            return list(data)
        except (TransportClosed, RuntimeError):
            logging.debug("PersistentDatagramStream: error on read")
            self._stream = None
        return None

    async def send(self, payload: list[int]) -> bool:
        """Sends message to stream"""
        if self._stream is None:
            return False
        try:
            await self._stream.send(bytes(payload))
            logging.debug("PersistentDatagramStream: sended data %s", list(payload))
            return True
        except (TransportClosed, RuntimeError):
            logging.debug("PersistentDatagramStream: error on write")
            self._stream = None
        return False



from .bytes import assert_byte
from .constants import (
    FIRST_HEADER_CODE,
    MAX_PAYLOAD_LENGTH,
    MIN_MESSAGE_LENGTH,
    SECOND_HEADER_CODE,
    SUCCESS_CODE,
    TAIL_CODE,
)


class Message:
    _feature: int
    _action: int
    _payload: bytes

    def __init__(self, feature: int, action: int, payload: bytes = None):
        assert_byte(feature)
        assert_byte(action)
        if payload is not None and len(payload) > MAX_PAYLOAD_LENGTH:
            raise ValueError("Payload is too long")
        self._feature = feature
        self._action = action
        if payload is None:
            payload = []
        self._payload = bytes(payload)

    def format_bytes(self) -> bytes:
        """Formats message bytes"""
        return bytes([
            FIRST_HEADER_CODE,
            SECOND_HEADER_CODE,
            len(self._payload)+2,
            self._feature,
            self._action,
            *self._payload,
            TAIL_CODE
        ])

    @property
    def feature(self) -> int:
        """Message feature code"""
        return self._feature

    @property
    def action(self) -> int:
        """Message action code"""
        return self._action

    @property
    def payload(self) -> bytes:
        """Message payload"""
        return self._payload

    @property
    def payload_without_status(self) -> bytes:
        return self._payload[1:]

    @property
    def is_successful(self) -> bool:
        """Is message successful"""
        return len(self._payload) > 0 and self._payload[0] == SUCCESS_CODE

def parse_message(message: bytes) -> Message:
    """Parse message from bytes"""
    if len(message) < 6:
        raise ValueError("Not enough bytes to parse message")
    if message[0] != FIRST_HEADER_CODE or message[1] != SECOND_HEADER_CODE:
        raise ValueError("Invalid header")
    payload_length = message[2] - 2
    if len(message) != MIN_MESSAGE_LENGTH + payload_length:
        raise ValueError("Not enough bytes to parse message")
    return Message(
        feature=message[3],
        action=message[4],
        payload=message[5:5+payload_length],
    )



"""MyrtIO byte utils"""

def split_byte_pair(value: int) -> tuple[int, int]:
    """Splits number into high and low bytes"""
    return value >> 8, value & 0xff

def from_byte_pair(high: int, low: int) -> int:
    """Assembles a number from high and low bytes."""
    return (high << 8) + low

def assert_byte(value: int):
    """Asserts that number is in range from 0 to 255"""
    assert 0 <= value <= 255
